#
# coding=utf-8
#*********************************************************************
#
# PyMPDTouchGui
# Gui for controlling the MusicPlayDaemon on Touchscreens
#
# written by Timo Boettcher
# pympdtouchgui (near) spida (.) net
#
# (C) 2007-2009 Timo Boettcher
# This is free software. You may redistribute copies of it under the
# terms of the GNU General Public License (Version 2)
# <http://www.gnu.org/licenses/gpl.html>.
# There is NO WARRANTY, to the extent permitted by law.
#
#*********************************************************************


try:
	import sys
	import os
	import string
	import logging
	import pygame
	from pygame import image
	import pygame.locals
	import events
except ImportError, err:
	print "couldn't load module. %s"%(err)
	sys.exit(2)


#==============================================================================
class Screen:
	"""Window managing object"""

	def __init__(self, Size, Config):
		self.Size = Size
		self.Config = Config
		self.Windows = []

	def Resize(self, newSize):
		(FactorX, FactorY) = (newSize[0]/self.Size[0], newSize[1]/self.Size[1])
		self.Size = newSize
		for Window in self.Windows:
			Window.ResizeFactor((FactorX, FactorY))

	def Paint(self):
		for Window in self.Windows:
			Window.Paint()

	def Clicked(self, Position):
		for Window in self.Windows:
			if Window.Position[0] < Position[0] and Position[0] < Window.Position[0] + Window.Size[0] and Window.Position[1] < Position[1] and Position[1] < Window.Position[1] + Window.Size[1]:
				Window.Clicked(Position)

#==============================================================================
class Window:
	"""Sprite managing object"""

	def __init__(self, Parent, Position, Size, Config):
		self.Parent = Parent
		self.Position = Position
		self.Size = Size
		self.rect = pygame.Rect(Position, Size)
		self.Config = Config
		self.Widgets = []
		self.Windows = []

		self.logger = logging.getLogger('pympdtouchgui')

	def Move(self, newPosition):
		# Calculate Difference between old and new Position
		(DiffX, DiffY) = (newPosition[0]-self.Position[0], newPosition[1]-self.Position[1])
		self.Position = newPosition
		# And apply it to all Child-Windows
		for Window in self.Windows:
			Window.Move((Window.rect.topleft[0]+DiffX, Window.rect.topleft[1]+DiffY))
		for Widget in self.Widgets:
			Widget.Move((Widget.rect.topleft[0]+DiffX, Widget.rect.topleft[1]+DiffY))

	def Resize(self, newSize):
		# Calculate Difference between old and new Size
		(FactorX, FactorY) = (float(newSize[0])/float(self.Size[0]), float(newSize[1])/float(self.Size[1]))
		self.Size = (int(newSize[0]),int(newSize[1]))
		# And apply it to all Child-Windows
		for Window in self.Windows:
			Window.ResizeFactor((FactorX, FactorY))
		for Widget in self.Widgets:
			Widget.ResizeFactor((FactorX, FactorY))

	def ResizeFactor(self, Factor):
		# Calculate Difference between old and new Size
		(FactorX, FactorY) = Factor
		if (FactorX <0 or FactorY<0):
			self.logger.fatal("Resizing by a negative Factor must be wrong")
			sys.exit(99)
		self.Size = (int(self.Size[0]*FactorX),int(self.Size[1]*FactorY))
		# And apply it to all Child-Windows
		for Window in self.Windows:
			Window.ResizeFactor((FactorX, FactorY))
		for Widget in self.Widgets:
			Widget.ResizeFactor((FactorX, FactorY))

	def Paint(self):
		for Window in self.Windows:
			Window.Paint()
		for Widget in self.Widgets:
			Widget.Paint()

	def Clicked(self, ClickX, ClickY):
		if self.Position[0] < ClickX and ClickX < self.Position[0] + self.Size[0] and self.Position[1] < ClickY and ClickY < self.Position[1] + self.Size[1]:
			for Window in self.Windows:
				if Window.rect.left < ClickX and ClickX < Window.rect.left + Window.rect.width and Window.rect.top < ClickY and ClickY < Window.rect.top + Window.rect.height:
					Window.Clicked(ClickX,ClickY)
			for Widget in self.Widgets:
				if Widget.rect.left < ClickX and ClickX < Widget.rect.left + Widget.rect.width and Widget.rect.top < ClickY and ClickY < Widget.rect.top + Widget.rect.height:
					Widget.Clicked(ClickX,ClickY)

	def Children(self):
		allchildren = []
		for Widget in self.Widgets:
			allchildren.append(Widget)
		for Window in self.Windows:
			for Widget in Window.Children():
				allchildren.append(Widget)
		return allchildren

	def kill(self):
		for Window in self.Windows:
			Window.kill()
		for Window in self.Windows:
			del(Window)
		for Widget in self.Widgets:
			Widget.kill()
		for Widget in self.Widgets:
			del(Widget)

#==============================================================================
class Widget(pygame.sprite.Sprite):
	"""Sprite with some general properties"""

	def __init__(self, Parent, Position, Size, Config, Data=""):
		pygame.sprite.Sprite.__init__(self)
		self.Parent = Parent
		self.Config = Config
		self.Tag = None
		self.image = pygame.Surface(Size) #FIXME sometimes called with floats, find out when and fix it
		self.rect = self.image.get_rect()
		self.rect.topleft = Position
		self.Paint()

	def Move(self, newPosition):
		#Use Resize before move
		self.rect.topleft = newPosition

	def Resize(self, newSize):
		#Use Resize before move
		self.image = pygame.Surface(newSize)
		self.rect = self.image.get_rect()
		self.Paint()

	def ResizeFactor(self, Factor):
		(FactorX, FactorY) = Factor
		if (FactorX <0 or FactorY<0):
			self.logger.fatal("Resizing by a negative Factor must be wrong")
			sys.exit(99)
		self.Resize((int(self.rect.width * FactorX), int(self.rect.height * FactorY)))

	def set_Data(self, Data):
		# defined in the implementations of this interface...
		pass

	def set_Tag(self, Tag):
		"""Field to hold Data that may be used on Callback"""
		self.Tag = Tag

	def Paint(self):
		# defined in the implementations of this interface...
		pass

	def Clicked(self, ClickX, ClickY):
		# defined in the implementations of this interface...
		pass

#==============================================================================
class ScrollText(Widget):
	Aligns = ["left","center","right"]
	def __init__(self, Parent, Position, Size, Config, FontSize=10, Color=(255,255,255)):
		self.Text = ""
		self.TextScrollIndex = 0
		self.TextWidth = 0
		self.TextScroll = False
		self.TextScrollDirectionForward = True
		self.TextScrollDirectionForwardOnly = True
		self.TextScrollSpeed = 2
		self.Align = "center"

		self.textImg = pygame.Surface(Size)
		self.FontSize = FontSize
		self.ForegroundColor = Color
		self.BackgroundColor = Config.Theme.ColorBackground
		self.changedsincepaint = True

		Widget.__init__(self, Parent, Position, Size, Config)


	def set_Data(self, Data):
		self.set_Text(Data)

	def set_Text(self, Text):
		self.Text = Text
		self.TextScrollIndex = 0

		font = pygame.font.Font(self.Config.FontName, self.FontSize)
		self.textImg = font.render(self.Text, self.Config.FontAntialias, self.ForegroundColor)
		self.TextWidth = self.textImg.get_width()

		self.changedsincepaint = True
		self.Paint()

	def Scroll(self):
		if self.TextScroll and self.TextWidth > self.rect.width:
			if self.TextScrollDirectionForward:
				self.TextScrollIndex += self.TextScrollSpeed
				if self.TextScrollIndex + self.rect.width > self.textImg.get_width():
					if self.TextScrollDirectionForwardOnly:
						self.TextScrollIndex = 0
					else:
						self.TextScrollIndex = self.textImg.get_width() - self.rect.width
						self.TextScrollDirectionForward = False
			else:
				self.TextScrollIndex -= self.TextScrollSpeed
				if self.TextScrollIndex <= 0:
					self.TextScrollIndex = 0
					self.TextScrollDirectionForward = True

		self.changedsincepaint = True
		self.Paint()


	def Paint(self):
		if self.changedsincepaint:
			self.changedsincepaint = False
			self.image.fill(self.BackgroundColor)
			if self.Align == "left":
				self.image.blit(self.textImg, (self.rect.left, self.rect.height/2 - self.textImg.get_height()/2 ), (self.TextScrollIndex, 0, self.rect.width, self.rect.height))
			elif self.Align == "center":
				self.image.blit(self.textImg, (self.rect.width/2 - min(self.textImg.get_width(),self.rect.width)/2, self.rect.height/2 - self.textImg.get_height()/2 ), (self.TextScrollIndex, 0, self.rect.width, self.rect.height))
			elif self.Align == "right":
				self.image.blit(self.textImg, (self.rect.right - self.textImg.get_width(), self.rect.height/2 - self.textImg.get_height()/2 ), (self.TextScrollIndex, 0, self.rect.width, self.rect.height))
			else:
				self.logger.fatal("Unknown Align")
				sys.exit(99)

	def update(self):
		self.Scroll()

#==============================================================================
class TextInput(Widget):
	Aligns = ["left","center","right"]
	def __init__(self, Parent, Position, Size, Config, FontSize=10):
		self.Text = ""
		self.TextScrollIndex = 0
		self.CallBack = None
		self.Align = "left"
		self.Blink = False
		self.Border = True
		self.FontSize = FontSize
		self.changedsincepaint = True

		Widget.__init__(self, Parent, Position, Size, Config)

	def set_Data(self, Data):
		self.set_Text(Data)

	def set_Text(self, Text):
		self.Text = Text
		self.TextScrollIndex = 0
		self.changedsincepaint = True
		self.Paint()

	def AddText(self, Text):
		self.set_Text(self.Text + Text)

	def Backspace(self):
		self.set_Text(self.Text[:-1])

	def HandleInput(self, Input):
		if isinstance(event, KeyboardEvent):
			if pygame.key.name(event.key) == "backspace":
				self.Text = self.Text[:-1]
			elif pygame.key.name(event.key) == "return":
				self.Callback(self.Text)
			elif pygame.key.name(event.key) in string.printable:
				self.Text = self.Text + pygame.key.name(event.key)
		self.changedsincepaint = True

	def set_Callback(self, Callback):
		self.Callback = Callback

	def Paint(self):
		if self.changedsincepaint:
			self.changedsincepaint = False
			self.image.fill(self.Config.Theme.ColorBackground)
			font = pygame.font.Font(self.Config.FontName, self.FontSize)
			if self.Blink:
				textImg = font.render(self.Text + "_", self.Config.FontAntialias, self.Config.Theme.ColorForegroundSearchText)
			else:
				textImg = font.render(self.Text, self.Config.FontAntialias, self.Config.Theme.ColorForegroundSearchText)

			if self.Border:
				pygame.draw.rect(self.image, self.Config.Theme.ColorForegroundSearchText, (1, self.rect.height/2 - (textImg.get_height()*1.5)/2, self.rect.width-1, textImg.get_height()*1.5), 2)

			Borderfactor = 50 # 2 Percent emptyness between border and text

			if self.Align == "left":
				self.image.blit(textImg, (int(self.rect.width/Borderfactor), self.rect.height/2 - textImg.get_height()/2 ))
			elif self.Align == "center":
				self.image.blit(textImg, (self.rect.width/2 - textImg.get_width()/2, self.rect.height/2 - textImg.get_height()/2 ))
			elif self.Align == "right":
				self.image.blit(textImg, (self.rect.width - textImg.get_width - int(self.rect.width/Borderfactor), self.rect.height/2 - textImg.get_height()/2 ))
			else:
				self.logger.fatal("Unknown Align")
				sys.exit(99)

	def BlinkCursor(self):
		self.Blink = not self.Blink
		self.Paint()

#==============================================================================
class Button:
	CallBack = None

	def set_Callback(self, Callback):
		self.CallBack = Callback

	def Clicked(self, ClickX, ClickY):
		if self.rect.left < ClickX and ClickX < self.rect.left + self.rect.width and self.rect.top < ClickY and ClickY < self.rect.top + self.rect.height:
			if self.CallBack != None:
				if self.Tag != "" and self.Tag != None:
					self.CallBack(self.Tag)
				else:
					self.CallBack()

#==============================================================================
class Image(Widget):
	def __init__(self, Parent, Position, Size, Config, Image=None):
		self.ImagePath = ""
		self.Img = pygame.Surface(Size)
		Widget.__init__(self, Parent, Position, Size, Config)
		if Image != None and Image != "":
			self.set_Image(Image)

	def set_Data(self, Data):
		self.set_Image(Data)

	def set_Image(self, ImagePath):
		self.Img.fill(self.Config.Theme.ColorBackground)
		if ImagePath != "" and ImagePath != None:
			ImagePath = ImagePath.encode("utf-8")
			try:
				if os.path.isfile(ImagePath):
					self.ImagePath = ImagePath
					self.Img = pygame.image.load(ImagePath)
					self.Img.convert()
				else:
					self.logger.warn("File `%s' does not exist"%ImagePath)
			except IOError, (errno, strerror):
				self.logger.warn("Error Opening %s: %s"%(ImagePath,strerror))
			except pygame.error, strerror:
				self.logger.warn("Error Opening %s: %s"%(ImagePath,strerror))
			except:
				self.logger.warn("Error Opening %s"%(ImagePath))
		self.Paint()

	def Paint(self):
		self.image.fill(self.Config.Theme.ColorBackground)
		ImgScaled = pygame.transform.scale(self.Img, (self.rect.width, self.rect.height))
		self.image.blit(ImgScaled, (0,0))

#==============================================================================
class BlackBox(Widget):
	def __init__(self, Parent, Position, Size, Config, Color=(0,0,0)):
		self.Img = pygame.Surface(Size)
		self.set_Color(Color)
		Widget.__init__(self, Parent, Position, Size, Config)

	def set_Data(self, Data):
		self.set_Color(Data)

	def set_Color(self, Color):
		self.Color = Color

	def Paint(self):
		self.image.fill(self.Color)

#==============================================================================
class ImageButton(Button, Image):
	def __init__(self, Parent, Position, Size, Config, ButtonType=""):
		Image.__init__(self, Parent, Position, Size, Config)
		self.ButtonType = ButtonType
		if ButtonType !="" and ButtonType != None:
			self.set_ButtonType(ButtonType)
		else:
			#FIXME do something
			pass

	def set_Data(self, Data):
		self.set_ButtonType(Data)

	def set_ButtonType(self, ButtonType):
		filename = os.path.join(self.Config.ThemeDir,"Button"+ButtonType+".png")
		self.ButtonType = ButtonType
		self.set_Image(filename)

#==============================================================================
class TextButton(Button, ScrollText):
	def __init__(self, Parent, Position, Size, Config, FontSize=10, Color=(255,255,255)):
		ScrollText.__init__(self, Parent, Position, Size, Config, FontSize, Color)

#==============================================================================
class TextBorderButton(Button, Widget):
	def __init__(self, Parent, Position, Size, Config, FontSize=10):
		self.Img = pygame.Surface(Size)
		self.FontSize = FontSize
		self.Text = ""
		self.ForegroundColor = (255,255,255)
		Widget.__init__(self, Parent, Position, Size, Config)

	def set_Data(self, Data):
		self.set_Text(Data)

	def set_Text(self, Text):
		self.image.fill(self.Config.Theme.ColorBackground)
		self.Text = Text
		ImagePath = os.path.join(self.Config.ThemeDir,"ButtonKey.png")
		try:
			if os.path.isfile(ImagePath):
				self.ImagePath = ImagePath
				self.Img = pygame.image.load(ImagePath)
				self.Img.convert()
				self.Paint()
			else:
				self.logger.warn("File `%s' does not exist"%ImagePath)
		except IOError, (errno, strerror):
			self.logger.warn("Error Opening %s: %s"%(filename,strerror))
		except pygame.error, strerror:
			self.logger.warn("Error Opening %s: %s"%(filename,strerror))
		except:
			self.logger.warn("Error Opening %s"%(filename))

	def Paint(self):
		self.image.fill(self.Config.Theme.ColorBackground)
		ImgScaled = pygame.transform.scale(self.Img, (self.rect.width, self.rect.height))
		self.image.blit(ImgScaled, (0,0))
		font = pygame.font.Font(self.Config.FontName, self.FontSize)
		self.textImg = font.render(self.Text, self.Config.FontAntialias, self.ForegroundColor)
		self.image.blit(self.textImg, (self.rect.width/2 - min(self.textImg.get_width(),self.rect.width)/2, self.rect.height/2 - self.textImg.get_height()/2 ))

#==============================================================================
class ScrollSlider(Button):
	def __init__(self, Parent, Position, Size, Config, ButtonType=""):
		Button.__init__(self, Parent, Position, Size, Config)
		self.ButtonType = ButtonType

#==============================================================================
class VirtualKeyboard(Window):
	def __init__(self, Parent, Position, Size, Config, Fontsize):
		Window.__init__(self, Parent, Position, Size, Config)
		self.Callback = None
		self.CallbackAddText = None
		self.CallbackBackspace = None

		self.Fontsize = Fontsize

	def set_Callback(self, Callback):
		self.Callback = Callback

	def set_CallbackAddText(self, Callback):
		self.CallbackAddText = Callback

	def set_CallbackBackspace(self, Callback):
		self.CallbackBackspace = Callback


#==============================================================================
class VirtualKeyboardKeyboard(VirtualKeyboard):
	def __init__(self, Parent, Position, Size, Config, Fontsize = 10):
		VirtualKeyboard.__init__(self, Parent, Position, Size, Config, Fontsize)

		ButtonsLine0 = [" "," "," "," "," "," "," "," ","Backspace","Enter"]
		ButtonsLine1 = ["Q","W","E","R","T","Z","U","I","O","P"]
		ButtonsLine2 =   ["A","S","D","F","G","H","J","K","L"]
		ButtonsLine3 = [" ","Y","X","C","V","B","N","M"," "," "]
		AllButtonsLines = [ButtonsLine0,ButtonsLine1,ButtonsLine2,ButtonsLine3]

		maxlinelen = 0
		for line in AllButtonsLines:
			if len(line) > maxlinelen:
				maxlinelen = len(line)
		ButtonWidth = int(Size[0] / maxlinelen)
		ButtonHeight = int(Size[1] / (len(AllButtonsLines)))
		ButtonSize = int(min(ButtonWidth,ButtonHeight) * 0.9)
		for linenumber,line in enumerate(AllButtonsLines):
			top = Position[1] + ButtonHeight * linenumber
			if len(line)  == maxlinelen:
				left = Position[0]
			else:
				left = Position[0] + int(ButtonWidth * (maxlinelen-len(line))/2)
			for ColumnNumber,ButtonCaption in enumerate(line):
				if ButtonCaption != " ":
					if len(ButtonCaption)<=1:
						spriteButton = TextBorderButton(self, (left,top), (ButtonSize,ButtonSize), Config, self.Fontsize)
						spriteButton.set_Callback(self.CallbackButton)
						spriteButton.set_Text(ButtonCaption)
						spriteButton.set_Tag(ButtonCaption)
					else:
						spriteButton = ImageButton(self, (left,top), (ButtonSize,ButtonSize), Config, ButtonCaption)
						spriteButton.set_Callback(self.CallbackButton)
						spriteButton.set_Tag(ButtonCaption)
					self.Widgets.append(spriteButton)
				left = left + ButtonWidth

	def CallbackButton(self, Tag):
		if len(Tag) == 1:
			if self.CallbackAddText:
				self.CallbackAddText(Tag)
		else:
			if Tag == "Backspace":
				if self.CallbackBackspace:
					self.CallbackBackspace()
			elif Tag == "Enter":
				if self.Callback:
					self.Callback()

#==============================================================================
class VirtualKeyboardStar(VirtualKeyboard):
	def __init__(self, Parent, Position, Size, Config, Fontsize = 10):
		VirtualKeyboard.__init__(self, Parent, Position, Size, Config, Fontsize)
		
		self.Fontsize = int(Fontsize * 3/4)
		font = pygame.font.Font(self.Config.FontName, self.Fontsize)

		self.GroupSelected = None

		self.Button = {}

		CharGroup0=["A","B","C","D"]
		CharGroup1=["E","F","G","H"]
		CharGroup2=["I","J","K","L"]
		CharGroup3=["M","N","O","P","Q"]
		CharGroup4=["R","S","T","U"]
		CharGroup5=["V","W","X","Y","Z"]
		self.CharGroup=[CharGroup0,CharGroup1,CharGroup2,CharGroup3,CharGroup4,CharGroup5]
		#self.CharGroup=[CharGroup0,CharGroup1,CharGroup2,CharGroup3,CharGroup4]

		# draw the buttons
		ButtonCount = len(self.CharGroup)
		RowCount = 2
		ColumnCount = int(ButtonCount/float(RowCount) + 0.5)


		maxwidth = 0
		for CharGroup in self.CharGroup:
			textImg = font.render(" ".join(CharGroup), self.Config.FontAntialias, (123,123,123))
			if textImg.get_width() > maxwidth:
				maxwidth = textImg.get_width()
	
		ButtonWidth = int(maxwidth * 1.4)
		ButtonHeight = int(font.get_linesize() * 1.5)
		ButtonSize = (ButtonWidth,ButtonHeight)

		rowtop = {}
		for RowNumber in xrange(0,RowCount):
			rowtop[RowNumber] = self.Position[1] + int(self.Size[1]/(RowCount+1)*(RowNumber+1))
		usedspace = ColumnCount * ButtonWidth
		leftstep = int((self.Size[0]-usedspace)/(ColumnCount + 1))
		leftstart = self.Position[0] + leftstep + int(ButtonSize[0]/2)

		currentrow = 0
		for ButtonNumber in xrange(0,ButtonCount):
			if ButtonNumber < ButtonCount:
				left = leftstart + (ButtonNumber % ColumnCount) * (leftstep+ButtonWidth) - int(ButtonSize[0]/2)
				top = rowtop[currentrow]
				self.Button[ButtonNumber] = TextBorderButton(self, (left,top), ButtonSize, self.Config, self.Fontsize)
				text = " ".join(self.CharGroup[ButtonNumber])
				self.Button[ButtonNumber].set_Text(text)
				self.Button[ButtonNumber].set_Tag(ButtonNumber)
				self.Widgets.append(self.Button[ButtonNumber])
				if ButtonNumber % ColumnCount >= ColumnCount - 1:
					currentrow = currentrow + 1


	def CallbackIR(self, Button):
		ButtonToGroup = {"1": 0, "2": 1, "3": 2, "4": 3, "5": 4, "6": 5, "7": 6}
		if ButtonToGroup.has_key(Button):
			if self.GroupSelected == None:
				# First step of selection: Select a group of chars
				if ButtonToGroup[Button] < len(self.CharGroup):
					self.GroupSelected = ButtonToGroup[Button]
					self.HighlightSelect(self.Button[self.GroupSelected])
			else:
				# Second step of selection: Select one char of the group
				if ButtonToGroup[Button] < len(self.CharGroup[self.GroupSelected]):
					Char = self.CharGroup[self.GroupSelected][ButtonToGroup[Button]]
					self.CallbackAddText(Char)
					self.HighlightSelect(self.Button[self.GroupSelected], False)
					self.GroupSelected = None
		elif Button == "Delete":
			if self.CallbackBackspace:
				self.CallbackBackspace()
		elif Button == "Cancel":
			if self.GroupSelected != None:
				self.HighlightSelect(self.Button[self.GroupSelected], False)
				self.GroupSelected = None

	def HighlightSelect(self, widget, highlight=True):
		if highlight:
			widget.BackgroundColor = self.Config.Theme.ColorBackgroundVirtualKeyboardSelect
			widget.ForegroundColor = self.Config.Theme.ColorForegroundVirtualKeyboardSelect
		else:
			widget.BackgroundColor = self.Config.Theme.ColorBackground
			widget.ForegroundColor = self.Config.Theme.ColorForegroundVirtualKeyboardText
		widget.set_Text(widget.Text)
		widget.Paint()

#==============================================================================
class ScrollableList(Window):
	def __init__(self, Parent, Position, Size, Config, Fontsize, RowFormat):
		Window.__init__(self, Parent, Position, Size, Config)
		self.Fontsize = Fontsize
		self.ScrollIndex = 0
		self.RowFormat = RowFormat
		font = pygame.font.Font(self.Config.FontName, self.Fontsize)
		self.RowHeight = int(font.get_linesize() * 1.3) + 6 # 3 Pixel Border
		self.ScrollbarWidth = self.RowHeight
		self.RowWidth = self.Size[0] - self.ScrollbarWidth
		self.RowCount = int(self.Size[1]/self.RowHeight)


		self.Data = []
		self.LineWidgets = {}
		self.CallbackScroll = None
		self.CallbackManualScroll = None

		#create scrollbar and connect
		self.Scrollbar = Scrollbar(self, (Position[0]+Size[0]-self.ScrollbarWidth,Position[1]), (self.ScrollbarWidth,Size[1]), Config)
		self.Scrollbar.UpdateData(0, self.RowCount)
		self.Scrollbar.set_CallbackChange(self.ScrollbarCallbackChange)
		self.Scrollbar.set_CallbackManualChange(self.ScrollbarCallbackManualChange)
		self.Windows.append(self.Scrollbar)


	def ScrollbarCallbackChange(self, current):
		"""Called when scrollbar was used to scroll"""
		self.ScrollToIndex(current)
		if self.CallbackScroll != None:
			self.CallbackScroll()

	def ScrollbarCallbackManualChange(self):
		"""Called when scrollbar was used to manually scroll"""
		if self.CallbackManualScroll != None:
			self.CallbackManualScroll()

	def set_CallbackScroll(self, Callback):
		self.CallbackScroll = Callback

	def set_CallbackManualScroll(self, Callback):
		self.CallbackManualScroll = Callback

	def Move(self, newPosition):
		Window.Move(self, Position, Size)

	def Resize(self, newSize):
		font = pygame.font.Font(self.Config.FontName, self.Fontsize)
		self.RowCount = int(newSize[1]/self.RowHeight)

		for LineNumber in xrange(self.RowCount,max(self.LineWidgets.keys())):
			if self.LineWidgets.has_key(LineNumber):
				for RowWidget in self.LineWidgets[LineNumber]:
					self.Widgets.remove(RowWidget)
					RowWidget.kill()
					RowWidget = None
					self.LineWidgets[LineNumber] = []

		Window.Resize(self, newSize)

		self.Scrollbar.UpdateScroll(self.Scrollbar.current)

	def set_Data(self, Data):
		self.Data = Data
		self.Scrollbar.UpdateScroll(0)
		self.Scrollbar.clear()
		self.Scrollbar.UpdateData(len(Data), self.RowCount)
		self.Paint()

	def ScrollRelative(self, offset):
		if offset < 0:
			if self.ScrollIndex + offset >= 0:
				self.ScrollIndex += offset
			else:
				self.ScrollIndex = 0
		elif offset > 0:
			if self.ScrollIndex + offset + self.RowCount < len(self.Data):
				self.ScrollIndex += offset
			else:
				self.ScrollIndex = len(self.Data) - self.RowCount
		self.Paint()

	def ScrollToIndex(self, index):
		if index >= 0:
			if self.RowCount < len(self.Data):
				# All Rows are filled
				if index + self.RowCount < len(self.Data):
					self.ScrollIndex = index
				else:
					# Tried to scroll below the last line: reset to last line
					self.ScrollIndex = len(self.Data) - self.RowCount
			else:
				# Not all Rows are filled
				self.ScrollIndex = 0
		else:
			# Invalid Position (before start): reset
			self.ScrollIndex = 0
		self.Scrollbar.current = self.ScrollIndex
		self.Paint()
		self.Scrollbar.ResizeScroll()

	def FindTag(self, searchcolumn, searchtag):
		resultrow = None
		for RowNumber,RowData in enumerate(self.Data):
			for ColumnNumber,ColumnData in enumerate(RowData):
				if ColumnNumber == searchcolumn:
					# got the right column
					if isinstance(ColumnData, tuple):
						# column has a tag
						if len(ColumnData)!=3:
							self.logger.fatal("Columndata is only allowed to have one tag and one Callback, but found %s tags"%len(ColumnData))
							sys.exit(99)
						else:
							Tag = ColumnData[2]
							if Tag == searchtag:
								# tag is the one we are looking for
								resultrow = RowNumber
		return resultrow

	def FindTagStart(self, searchcolumn, searchtag):
		resultrow = None
		for RowNumber,RowData in enumerate(self.Data):
			for ColumnNumber,ColumnData in enumerate(RowData):
				if ColumnNumber == searchcolumn:
					# got the right column
					if isinstance(ColumnData, tuple):
						# column has a tag
						if len(ColumnData)!=3:
							self.logger.fatal("Columndata is only allowed to have one tag and one Callback, but found %s tags"%len(ColumnData))
							sys.exit(99)
						else:
							Tag = ColumnData[2]
							if Tag.lower().startswith(searchtag.lower()):
								# tag is the one we are looking for
								# only find the first matching tag
								if resultrow == None or RowNumber < resultrow:
									resultrow = RowNumber
		return resultrow

	def WidgetInDataRow(self, searchcolumn, DataRowNumber):
		LineNumber = DataRowNumber - self.ScrollIndex
		if LineNumber < self.RowCount and LineNumber >= 0:
			# Widget is visible
			return self.LineWidgets[LineNumber][searchcolumn] # FIXME is this always in column-order?

	def WidgetWithTag(self, searchcolumn, searchtag):
		DataRowNumber = self.FindTag(searchcolumn, searchtag)
		if DataRowNumber != None:
			return self.WidgetInDataRow(searchcolumn, DataRowNumber)

	def Paint(self):
		LastIndex = min(self.ScrollIndex+self.RowCount,len(self.Data))
		for LineNumber,DataIndex in enumerate(xrange(self.ScrollIndex,LastIndex)):
			RowData = self.Data[DataIndex]
			self.PaintSingleRow(LineNumber,RowData)
		if LastIndex - self.ScrollIndex < self.RowCount:
			for LineNumber in xrange(LastIndex,self.RowCount):
				self.PaintSingleRow(LineNumber,None,True)


	def PaintSingleRow(self, LineNumber, RowData, Clear = False):
		if Clear:
			RowArray = []
			if self.LineWidgets.has_key(LineNumber):
				for RowWidget in self.LineWidgets[LineNumber]:
					self.Widgets.remove(RowWidget)
					RowWidget.kill()
					self.LineWidgets[LineNumber] = []

			Row = (BlackBox, self.Config.Theme.ColorBackground, None, None, self.RowWidth)
			RowArray.append(Row)

			self.LineWidgets[LineNumber] = self.DefaultRowGeneratorFunction((self.Position[0], self.Position[1]+LineNumber*self.RowHeight),self.RowHeight,RowArray)

			for RowWidget in self.LineWidgets[LineNumber]:
				self.Widgets.append(RowWidget)

		elif len(RowData) != len(self.RowFormat):
			self.logger.fatal("Format doesn't match Data. Expected %s Columns, but got %s."%(len(self.RowFormat),len(RowData)))
			sys.exit(99)
		else:
			RowArray = []
			for ColumnNumber,ColumnData in enumerate(RowData):
				Callback = None
				Tag = None
				if isinstance(ColumnData, tuple):
					if len(ColumnData)!=3:
						self.logger.fatal("Columndata is only allowed to have one tag and one Callback, but found %s tags"%len(ColumnData))
						sys.exit(99)
					else:
						Tag = ColumnData[2]
						Callback = ColumnData[1]
						ColumnData = ColumnData[0]
				(Type, Width) = self.RowFormat[ColumnNumber]
				Column = (Type, ColumnData, Callback, Tag, Width)
				RowArray.append(Column)

			if self.LineWidgets.has_key(LineNumber):
				for RowWidget in self.LineWidgets[LineNumber]:
					self.Widgets.remove(RowWidget)
					RowWidget.kill()
					self.LineWidgets[LineNumber] = []

			self.LineWidgets[LineNumber] = self.DefaultRowGeneratorFunction((self.Position[0], self.Position[1]+LineNumber*self.RowHeight),self.RowHeight,RowArray)

			for RowWidget in self.LineWidgets[LineNumber]:
				self.Widgets.append(RowWidget)

	# Could be anonymous
	def DefaultRowGeneratorFunction(self, Position, Height, RowArray):
		LeftPos = 0
		RowWidgets = []

		for ColumNumber,ColumnData in enumerate(RowArray):
			(Type, Data, Callback, Tag, Width) = ColumnData
			if Type == ScrollText or Type == TextButton:
				ColumnWidget = Type(self, (Position[0] + LeftPos, Position[1]), (Width, Height), self.Config, self.Fontsize)
			else:
				ColumnWidget = Type(self, (Position[0] + LeftPos, Position[1]), (Width, Height), self.Config)
			if Tag != "" and Tag != None:
				ColumnWidget.set_Tag(Tag)
			if isinstance(ColumnWidget,Button) and Callback:
				ColumnWidget.set_Callback(Callback)

			ColumnWidget.set_Tag = Tag
			if Data != "" and Data != None:
				ColumnWidget.set_Data(Data)
			RowWidgets.append(ColumnWidget)
			LeftPos = LeftPos + Width
		return RowWidgets

#==============================================================================
class Scrollbar(Window):
	def __init__(self, Parent, Position, Size, Config, DirectionHoriz=False):
		Window.__init__(self, Parent, Position, Size, Config)
		self.Config = Config
		self.DirectionHoriz = DirectionHoriz
		self.count = 0
		self.current = 0
		self.visible = 0
		self.Elements = {}
		self.CallbackChange = None
		self.CallbackManualChange = None

		# Top Button
		if self.DirectionHoriz:
			self.ButtonSize = self.Size[1]
			size = (self.ButtonSize,self.ButtonSize)
			position = self.Position
			FirstButton = ImageButton(self, position, size, self.Config, "ScrollLeft")
		else:
			self.ButtonSize = self.Size[0]
			size = (self.ButtonSize,self.ButtonSize)
			position = self.Position
			FirstButton = ImageButton(self, position, size, self.Config, "ScrollUp")
		FirstButton.set_Callback(self.CallbackScrollUp)
		self.Elements["FirstButton"] = FirstButton
		self.Widgets.append(FirstButton)

		# Space above Slider
		if self.DirectionHoriz:
			self.ButtonSize = self.Size[1]
			size = (self.ButtonSize,1)
			position = (self.Position[0]+self.ButtonSize,self.Position[1])
		else:
			self.ButtonSize = self.Size[0]
			size = (self.ButtonSize,1)
			position = (self.Position[0],self.Position[1]+self.ButtonSize)
		FirstSlider = ImageButton(self, position, size, self.Config, "FirstSlider")
		FirstSlider.set_Callback(self.CallbackSliderTop)
		self.Elements["FirstSlider"] = FirstSlider
		self.Widgets.append(FirstSlider)

		# Slider
		if self.DirectionHoriz:
			self.ButtonSize = self.Size[1]
			size = (self.ButtonSize,1)
			position = (self.Position[0]+self.ButtonSize+1,self.Position[1])
		else:
			self.ButtonSize = self.Size[0]
			size = (self.ButtonSize,1)
			position = (self.Position[0],self.Position[1]+self.ButtonSize+1)
		MiddleSlider = ImageButton(self, position, size, self.Config, "MiddleSlider")
		MiddleSlider.set_Callback(self.CallbackSliderMiddle)
		self.Elements["MiddleSlider"] = MiddleSlider
		self.Widgets.append(MiddleSlider)

		# Space below Slider
		if self.DirectionHoriz:
			self.ButtonSize = self.Size[1]
			size = (self.ButtonSize,1)
			position = (self.Position[0]+self.ButtonSize+2,self.Position[1])
		else:
			self.ButtonSize = self.Size[0]
			size = (self.ButtonSize,1)
			position = (self.Position[0],self.Position[1]+self.ButtonSize+2)
		SecondSlider = ImageButton(self, position, size, self.Config, "SecondSlider")
		SecondSlider.set_Callback(self.CallbackSliderBottom)
		self.Elements["SecondSlider"] = SecondSlider
		self.Widgets.append(SecondSlider)

		# Bottom Button
		if self.DirectionHoriz:
			self.ButtonSize = self.Size[1]
			size = (self.ButtonSize,self.ButtonSize)
			position = (self.Position[0]+self.Size[0]-self.ButtonSize,self.Position[1])
			SecondButton = ImageButton(self, position, size, self.Config, "ScrollRight")
		else:
			self.ButtonSize = self.Size[0]
			size = (self.ButtonSize,self.ButtonSize)
			position = (self.Position[0],self.Position[1]+self.Size[1]-self.ButtonSize)
			SecondButton = ImageButton(self, position, size, self.Config, "ScrollDown")
		SecondButton.set_Callback(self.CallbackScrollDown)
		self.Elements["SecondButton"] = SecondButton
		self.Widgets.append(SecondButton)



	def set_CallbackChange(self, CallbackChange):
		self.CallbackChange = CallbackChange

	def set_CallbackManualChange(self, CallbackManualChange):
		self.CallbackManualChange = CallbackManualChange

	def CallbackScrollUp(self):
		self.UpdateScroll(self.current - 1)
		if self.CallbackManualChange != None:
			self.CallbackManualChange()

	def CallbackSliderTop(self):
		self.UpdateScroll(self.current - self.visible)
		if self.CallbackManualChange != None:
			self.CallbackManualChange()

	def CallbackSliderMiddle(self):
		pass

	def CallbackSliderBottom(self):
		self.UpdateScroll(self.current + self.visible)
		if self.CallbackManualChange != None:
			self.CallbackManualChange()

	def CallbackScrollDown(self):
		self.UpdateScroll(self.current + 1)
		if self.CallbackManualChange != None:
			self.CallbackManualChange()

	def clear(self):
		self.count = 0
		self.current = 0
		self.visible = 0

	def UpdateData(self, count, visible):
		self.count = count
		self.visible = visible

		self.ResizeScroll()
		self.Write()

	def UpdateScroll(self, current):
		if self.count > self.visible:
			if current < 0:
				self.current = 0
			else:
				if current + self.visible > self.count:
					self.current = self.count - self.visible
				else:
					self.current = current
		else:
			self.current = 0

		self.ResizeScroll()
		self.Write()
		self.CallbackChange(self.current)

	def update(self):
		pass

	def ResizeScroll(self):
		if self.DirectionHoriz:
			Start2 = self.Position[0]
			Start = self.Position[0]
			Width = self.Size[1]
			TotalLength = self.Size[0] - Width * 2
		else:
			Start2 = self.Position[0]
			Start = self.Position[1]
			Width = self.Size[0]
			TotalLength = self.Size[1] - Width * 2

		if self.count > self.visible:
			MiddleSliderLength = int( float(self.visible * TotalLength) / float(self.count))
			FreeLength = TotalLength - MiddleSliderLength
			FirstSliderLength = int( float(self.current * FreeLength) / float(self.count - self.visible))
			SecondSliderLength = FreeLength - FirstSliderLength

			FirstSliderStart = Start + Width
			MiddleSliderStart = FirstSliderStart + FirstSliderLength
			SecondSliderStart = MiddleSliderStart + MiddleSliderLength
		else:
			FirstSliderStart = Start + Width
			FirstSliderLength = 1
			MiddleSliderStart = FirstSliderStart + FirstSliderLength
			MiddleSliderLength = TotalLength - 2
			SecondSliderStart = MiddleSliderStart + MiddleSliderLength
			SecondSliderLength = 1

		self.Elements["FirstButton"].Resize((self.ButtonSize,self.ButtonSize))
		self.Elements["FirstButton"].Move(self.Position)
		self.Elements["SecondButton"].Resize((self.ButtonSize,self.ButtonSize))
		if self.DirectionHoriz:
			self.Elements["FirstSlider"].Resize((FirstSliderLength,Width))
			self.Elements["FirstSlider"].Move((FirstSliderStart,Start2))
			self.Elements["MiddleSlider"].Resize((MiddleSliderLength,Width))
			self.Elements["MiddleSlider"].Move((MiddleSliderStart,Start2))
			self.Elements["SecondSlider"].Resize((LastSliderLength,Width))
			self.Elements["SecondSlider"].Move((LastSliderStart,Start2))
			self.Elements["SecondButton"].Move((self.Position[0]+self.Size[0]-self.ButtonSize,self.Position[1]))
		else:
			self.Elements["FirstSlider"].Resize((Width,FirstSliderLength))
			self.Elements["FirstSlider"].Move((Start2,FirstSliderStart))
			self.Elements["MiddleSlider"].Resize((Width,MiddleSliderLength))
			self.Elements["MiddleSlider"].Move((Start2,MiddleSliderStart))
			self.Elements["SecondSlider"].Resize((Width,SecondSliderLength))
			self.Elements["SecondSlider"].Move((Start2,SecondSliderStart))
			self.Elements["SecondButton"].Move((self.Position[0],self.Position[1]+self.Size[1]-self.ButtonSize))


	def Write(self):
		self.Elements["FirstSlider"].update()
		self.Elements["MiddleSlider"].update()
		self.Elements["SecondSlider"].update()

#==============================================================================
def maxTextWidth(FontSize, TextArray):
	maxwidth = 0
	font = pygame.font.Font(self.Config.FontName, Fontsize)
	for Text in self.TextArray:
		textImg = font.render(Text + "jM", self.Config.FontAntialias, (0,0,0)) # ":" and 1 em length, ":" substituted to accomodate for characters with sublength
		if textImg.get_width() > maxwidth:
			maxwidth = textImg.get_width()
	return maxwidth
